# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from varcode.effects.effect_classes import ExonicSpliceSite, Substitution
from cohorts import Cohort
from cohorts.variant_filters import no_filter
from cohorts.functions import *

from .data_generate import generate_vcfs

from nose.tools import eq_, ok_
from mock import MagicMock
from os import path
from shutil import rmtree
from collections import defaultdict
import pandas as pd

from .test_count import make_cohort

FILE_FORMAT_1 = "patient_format1_%s.vcf"

def test_default_filter_fn():
    """
    Test that filter_fn falls back to the Cohort default but can be overridden.
    """
    vcf_dir, cohort = None, None
    try:
        def default_filter_fn(filterable_variant):
            return filterable_variant.variant.start != 53513530

        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df["snv_count"]), [3, 3, 6])

        cohort.filter_fn = default_filter_fn
        df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df["snv_count"]), [2, 2, 5])

        df = cohort.as_dataframe(snv_count, filter_fn=None)
        eq_(list(df["snv_count"]), [2, 2, 5])

        df = cohort.as_dataframe(snv_count, filter_fn=no_filter)
        eq_(list(df["snv_count"]), [3, 3, 6])

        def another_filter_fn(filterable_variant):
            return default_filter_fn(filterable_variant) and filterable_variant.variant.start != 49658590
        df = cohort.as_dataframe(snv_count, filter_fn=another_filter_fn)
        eq_(list(df["snv_count"]), [1, 1, 4])
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_default_normalized_per_mb():
    """
    Test that normalized_per_mb falls back to the Cohort default but can be overridden.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        eq_(list(df["snv_count"]), [3, 3, 6])

        def load_ensembl_coverage(self):
            return pd.DataFrame({"patient_id": ["1", "4", "5"],
                                 "Num Loci": [2000000, 5000000, 8000000],
                                 "MB": [2, 5, 8]})
        Cohort.load_ensembl_coverage = load_ensembl_coverage
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])
        cohort.normalized_per_mb = True

        df = cohort.as_dataframe(snv_count)
        eq_(len(df), 3)
        # 3 / 2, 3 / 5, 6 / 8
        eq_(list(df["snv_count"]), [1.5, 0.6, 0.75])

        df = cohort.as_dataframe(snv_count, normalized_per_mb=None)
        eq_(list(df["snv_count"]), [1.5, 0.6, 0.75])

        df = cohort.as_dataframe(snv_count, normalized_per_mb=False)
        eq_(list(df["snv_count"]), [3, 3, 6])
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()

def test_default_filter_fn_multiple_functions():
    """
    Across multiple functions, test that filter_fn falls back to the Cohort default.
    """
    vcf_dir, cohort = None, None
    try:
        vcf_dir, cohort = make_cohort([FILE_FORMAT_1])

        def default_filter_fn(filterable_variant):
            return filterable_variant.variant.ref == "A"

        # Mock out load_single_patient_isovar since we don't actually have that data,
        # yet it's needed for expressed_missense_snv_count.
        variants = cohort.load_variants()
        isovar_mocked = {}
        for patient in cohort:
            chr_list = []
            pos_list = []
            ref_list = []
            alt_list = []
            patient_variants = variants[patient.id]
            for variant in patient_variants:
                chr_list.append(variant.contig)
                pos_list.append(variant.start)
                ref_list.append(variant.ref)
                alt_list.append(variant.alt)
            df_isovar_mocked = pd.DataFrame({"chr": chr_list,
                                             "pos": pos_list,
                                             "ref": ref_list,
                                             "alt": alt_list})
            isovar_mocked[patient.id] = df_isovar_mocked

        def mock_function(patient, **kwargs):
            return isovar_mocked[patient.id]

        cohort.load_single_patient_isovar = MagicMock(side_effect=mock_function)

        # Especially need to include expressed_missense_snv_count, which works a little differently.
        for count_func in [snv_count, missense_snv_count, expressed_missense_snv_count]:
            count_col = count_func.__name__

            # No Cohort default
            cohort.filter_fn = None
            df_no_arg  = cohort.as_dataframe(count_func)
            df_default_arg = cohort.as_dataframe(count_func, filter_fn=default_filter_fn)
            df_none_arg = cohort.as_dataframe(count_func, filter_fn=None)
            df_no_filter_arg = cohort.as_dataframe(count_func, filter_fn=no_filter)
            ok_((df_no_arg[count_col] == df_none_arg[count_col]).all())
            ok_((df_default_arg[count_col] != df_no_arg[count_col]).any())
            ok_((df_no_filter_arg[count_col] == df_no_arg[count_col]).all())

            # With a Cohort default
            cohort.filter_fn = default_filter_fn
            df_no_arg  = cohort.as_dataframe(count_func)
            df_default_arg = cohort.as_dataframe(count_func, filter_fn=default_filter_fn)
            df_none_arg = cohort.as_dataframe(count_func, filter_fn=None)
            df_no_filter_arg = cohort.as_dataframe(count_func, filter_fn=no_filter)
            ok_((df_no_arg[count_col] == df_none_arg[count_col]).all())
            ok_((df_default_arg[count_col] == df_no_arg[count_col]).all())
            ok_((df_no_filter_arg[count_col] != df_no_arg[count_col]).any())
    finally:
        if vcf_dir is not None and path.exists(vcf_dir):
            rmtree(vcf_dir)
        if cohort is not None:
            cohort.clear_caches()



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import pandas as pd
from os import path
from numpy.random import choice, randint, seed
from varcode import Variant, VariantCollection
from varcode.effects import Substitution
from mhctools.random_predictor import RandomBindingPredictor

from . import Cohort, Patient

def random_cohort(size, cache_dir, data_dir=None,
                  min_random_variants=None,
                  max_random_variants=None,
                  seed_val=1234):
    """
    Parameters
    ----------
    min_random_variants: optional, int
        Minimum number of random variants to be generated per patient.
    max_random_variants: optional, int
        Maximum number of random variants to be generated per patient.
    """
    seed(seed_val)
    d = {}
    d["id"] = [str(id) for id in range(size)]
    d["age"] = choice([10, 15, 28, 32, 59, 62, 64, 66, 68], size)
    d["smoker"] = choice([False, True], size)
    d["OS"] = [randint(10, 1000) for i in range(size)]
    d["PFS"] = [int(os * 0.6) for os in d["OS"]]
    d["benefit"] = [pfs < 50 for pfs in d["PFS"]]
    d["random"] = [randint(100) for i in range(size)]
    d["random_boolean"] = choice([False, True], size)
    d["benefit_correlate"] = [randint(50) if benefit else randint(20) for benefit in d["benefit"]]
    d["benefit_correlate_boolean"] = [True if corr > 10 else False for corr in d["benefit_correlate"]]
    d["deceased"] = choice([False, True], size)
    d["progressed_or_deceased"] = [deceased or choice([False, True]) for deceased in d["deceased"]]
    df = pd.DataFrame(d)
    patients = []
    for i, row in df.iterrows():
        snv_vcf_paths = None
        if max_random_variants is not None and min_random_variants is not None:
            if data_dir is None:
                raise ValueError("data_dir must be provided if random variants are being generated.")
            vcf_path = path.join(data_dir, "patient_%s_mutect.vcf" % row["id"])
            generate_simple_vcf(
                vcf_path, generate_random_missense_variants(num_variants=randint(min_random_variants, max_random_variants)))
            snv_vcf_paths = [vcf_path]
        patient = Patient(
            id=row["id"],
            os=row["OS"],
            pfs=row["PFS"],
            benefit=row["benefit"],
            deceased=row["deceased"],
            progressed_or_deceased=row["progressed_or_deceased"],
            hla_alleles=["HLA-A02:01"],
            variants={"snv": snv_vcf_paths},
            additional_data=row)
        patients.append(patient)
    return Cohort(
        patients=patients,
        cache_dir=cache_dir,
        mhc_class=RandomBindingPredictor)

def generate_random_missense_variants(num_variants=10, max_search=100000, reference="GRCh37"):
    """
    Generate a random collection of missense variants by trying random variants repeatedly.
    """
    variants = []
    for i in range(max_search):
        bases = ["A", "C", "T", "G"]
        random_ref = choice(bases)
        bases.remove(random_ref)
        random_alt = choice(bases)
        random_contig = choice(["1", "2", "3", "4", "5"])
        random_variant = Variant(contig=random_contig, start=randint(1, 1000000),
                                 ref=random_ref, alt=random_alt, ensembl=reference)
        try:
            effects = random_variant.effects()
            for effect in effects:
                if isinstance(effect, Substitution):
                    variants.append(random_variant)
                    break
        except:
            continue
        if len(variants) == num_variants:
            break
    return VariantCollection(variants)

def generate_simple_vcf(filename, variant_collection):
    """
    Output a very simple metadata-free VCF for each variant in a variant_collection.
    """
    contigs = []
    positions = []
    refs = []
    alts = []
    for variant in variant_collection:
        contigs.append("chr" + variant.contig)
        positions.append(variant.start)
        refs.append(variant.ref)
        alts.append(variant.alt)
    df = pd.DataFrame()
    df["contig"] = contigs
    df["position"] = positions
    df["id"] = ["."] * len(variant_collection)
    df["ref"] = refs
    df["alt"] = alts
    df["qual"] = ["."] * len(variant_collection)
    df["filter"] = ["."] * len(variant_collection)
    df["info"] = ["."] * len(variant_collection)
    df["format"] = ["GT:AD:DP"] * len(variant_collection)
    normal_ref_depths = [randint(1, 10) for v in variant_collection]
    normal_alt_depths = [randint(1, 10) for v in variant_collection]
    df["n1"] = ["0:%d,%d:%d" % (normal_ref_depths[i], normal_alt_depths[i],
                                normal_ref_depths[i] + normal_alt_depths[i])
                for i in range(len(variant_collection))]
    tumor_ref_depths = [randint(1, 10) for v in variant_collection]
    tumor_alt_depths = [randint(1, 10) for v in variant_collection]
    df["t1"] = ["0/1:%d,%d:%d" % (tumor_ref_depths[i], tumor_alt_depths[i], tumor_ref_depths[i] + tumor_alt_depths[i])
                for i in range(len(variant_collection))]

    with open(filename, "w") as f:
        f.write("##fileformat=VCFv4.1\n")
        f.write("##reference=file:///projects/ngs/resources/gatk/2.3/ucsc.hg19.fasta\n")

    with open(filename, "a") as f:
        df.to_csv(f, sep="\t", index=None, header=None)



# Copyright (c) 2016. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from .variant_filters import no_filter, effect_expressed_filter
from .utils import first_not_none_param

from functools import wraps
import numpy as np
from varcode.effects import Substitution
from varcode.common import memoize

def use_defaults(func):
    """
    Decorator for functions that should automatically fall back to the Cohort-default filter_fn and
    normalized_per_mb if not specified.
    """
    @wraps(func)
    def wrapper(row, cohort, filter_fn=None, normalized_per_mb=None, **kwargs):
        filter_fn = first_not_none_param([filter_fn, cohort.filter_fn], no_filter)
        normalized_per_mb = first_not_none_param([normalized_per_mb, cohort.normalized_per_mb], False)
        return func(row=row,
                    cohort=cohort,
                    filter_fn=filter_fn,
                    normalized_per_mb=normalized_per_mb,
                    **kwargs)
    return wrapper

def count_function(func):
    """
    Decorator for functions that return a collection (technically a dict of collections) that should be
    counted up. Also automatically falls back to the Cohort-default filter_fn and normalized_per_mb if
    not specified.
    """
    # Fall back to Cohort-level defaults.
    @use_defaults
    @wraps(func)
    def wrapper(row, cohort, filter_fn=None, normalized_per_mb=None, **kwargs):
        per_patient_data = func(row=row,
                                cohort=cohort,
                                filter_fn=filter_fn,
                                normalized_per_mb=normalized_per_mb,
                                **kwargs)
        patient_id = row["patient_id"]
        if patient_id in per_patient_data:
            count = len(per_patient_data[patient_id])
            if normalized_per_mb:
                count /= float(get_patient_to_mb(cohort)[patient_id])
            return count
        return np.nan
    return wrapper

@memoize
def get_patient_to_mb(cohort):
    patient_to_mb = dict(cohort.as_dataframe(join_with="ensembl_coverage")[["patient_id", "MB"]].to_dict("split")["data"])
    return patient_to_mb

@count_function
def snv_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    patient_id = row["patient_id"]
    return cohort.load_variants(
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=filter_fn,
        **kwargs)

@count_function
def nonsynonymous_snv_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    # This only loads one effect per variant.
    patient_id = row["patient_id"]
    return cohort.load_effects(
        only_nonsynonymous=True,
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=filter_fn,
        **kwargs)

@count_function
def missense_snv_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    def missense_filter_fn(filterable_effect):
        assert filter_fn is not None, "filter_fn should never be None, but it is."
        return (type(filterable_effect.effect) == Substitution and
                filter_fn(filterable_effect))
    # This only loads one effect per variant.
    patient_id = row["patient_id"]
    return cohort.load_effects(
        only_nonsynonymous=True,
        patients=[cohort.patient_from_id(patient_id)],
        filter_fn=missense_filter_fn,
        **kwargs)

@count_function
def neoantigen_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    patient = cohort.patient_from_id(row["patient_id"])
    return cohort.load_neoantigens(patients=[patient],
                                   filter_fn=filter_fn,
                                   **kwargs)

@use_defaults
def expressed_missense_snv_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    def expressed_filter_fn(filterable_effect):
        assert filter_fn is not None, "filter_fn should never be None, but it is."
        return filter_fn(filterable_effect) and effect_expressed_filter(filterable_effect)
    return missense_snv_count(row=row,
                              cohort=cohort,
                              filter_fn=expressed_filter_fn,
                              normalized_per_mb=normalized_per_mb)

@use_defaults
def expressed_neoantigen_count(row, cohort, filter_fn, normalized_per_mb, **kwargs):
    return neoantigen_count(row=row,
                            cohort=cohort,
                            filter_fn=filter_fn,
                            normalized_per_mb=normalized_per_mb,
                            only_expressed=True,
                            **kwargs)
